# Generated by Django 2.0.7 on 2018-07-25 18:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0003_staffdocument_public'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffprofile',
            name='birthday',
            field=models.DateField(
                blank=True, default=None, null=True, verbose_name='Birthday'),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-22 15:08

from django.db import migrations
import private_storage.fields
import private_storage.storage.files
import sorl.thumbnail.fields


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='staffprofile',
            name='image',
            field=sorl.thumbnail.fields.ImageField(
                blank=True,
                help_text='A square image works best',
                max_length=255,
                upload_to='staff-images/',
                verbose_name='Profile Image'),
        ),
        migrations.AlterField(
            model_name='staffdocument',
            name='file',
            field=private_storage.fields.PrivateFileField(
                help_text='Upload staff member document',
                storage=private_storage.storage.files.PrivateFileSystemStorage(
                ),
                upload_to='staff-documents/',
                verbose_name='File'),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-25 16:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0002_auto_20180722_1808'),
    ]

    operations = [
        migrations.AddField(
            model_name='staffdocument',
            name='public',
            field=models.BooleanField(
                default=False,
                help_text='If public, it will be available to everyone.',
                verbose_name='Public'),
        ),
    ]



"""
Apps module for small-small-hr
"""
from django.apps import AppConfig
from django.db.models.signals import post_migrate
from django.utils.translation import gettext_lazy as _


def mptt_callback(sender, **kwargs):  # pylint: disable=unused-argument
    """
    Rebuild mptt tree.

    We need to do this so that for existing objects we can rebuild the mptt tree
    and set correct values for all the mptt fields.  This is necessary because we
    need to remove the defaults placed in the migrations file.
    """
    # pylint: disable=import-outside-toplevel
    from small_small_hr.models import StaffProfile

    StaffProfile.objects.rebuild()


class SmallSmallHrConfig(AppConfig):
    """Apps config class."""

    name = "small_small_hr"
    app_label = "small_small_hr"
    verbose_name = _("Small Small HR")

    def ready(self):
        """Do stuff when the app is ready."""
        # pylint: disable=import-outside-toplevel,unused-import

        # post migrate function
        post_migrate.connect(mptt_callback, sender=self)

        # signals
        import small_small_hr.signals  # noqa

        # set up app settings
        from django.conf import settings
        import small_small_hr.settings as defaults

        for name in dir(defaults):
            if name.isupper() and not hasattr(settings, name):
                setattr(settings, name, getattr(defaults, name))



# Generated by Django 3.0.7 on 2020-06-26 10:13
# pylint: disable=invalid-name,missing-module-docstring,missing-class-docstring
import django.db.models.deletion
from django.db import migrations

import mptt.fields


class Migration(migrations.Migration):

    dependencies = [
        ("small_small_hr", "0009_auto_20200607_2244"),
    ]

    operations = [
        migrations.AlterField(
            model_name="staffprofile",
            name="supervisor",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="children",
                to="small_small_hr.StaffProfile",
                verbose_name="Manager",
            ),
        ),
    ]



"""Template tags module."""
from django import template
from django.utils.translation import ugettext as _

from small_small_hr.constants import HOURS, MINUTES

register = template.Library()


@register.filter
def overtime_duration(time_delta):
    """Display overtime duration."""
    total_seconds = int(time_delta.total_seconds())
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60

    return f"{hours} {_(HOURS)} {minutes} {_(MINUTES)}"



# Generated by Django 3.0.7 on 2020-06-07 12:37
# pylint: disable=invalid-name,missing-module-docstring,missing-class-docstring
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("small_small_hr", "0007_auto_20190625_2111"),
    ]

    operations = [
        migrations.RenameField(
            model_name="leave", old_name="reason", new_name="review_reason",
        ),
        migrations.RenameField(
            model_name="leave", old_name="status", new_name="review_status",
        ),
        migrations.RenameField(
            model_name="overtime", old_name="reason", new_name="review_reason",
        ),
        migrations.RenameField(
            model_name="overtime", old_name="status", new_name="review_status",
        ),
        migrations.RemoveField(model_name="leave", name="comments",),
        migrations.RemoveField(model_name="overtime", name="comments",),
        migrations.AddField(
            model_name="leave",
            name="review_date",
            field=models.DateTimeField(
                blank=True, default=None, null=True, verbose_name="Review Date"
            ),
        ),
        migrations.AddField(
            model_name="overtime",
            name="review_date",
            field=models.DateTimeField(
                blank=True, default=None, null=True, verbose_name="Review Date"
            ),
        ),
        migrations.AlterField(
            model_name="annualleave",
            name="year",
            field=models.PositiveIntegerField(
                choices=[
                    (2017, 2017),
                    (2018, 2018),
                    (2019, 2019),
                    (2020, 2020),
                    (2021, 2021),
                    (2022, 2022),
                    (2023, 2023),
                    (2024, 2024),
                    (2025, 2025),
                    (2026, 2026),
                    (2027, 2027),
                    (2028, 2028),
                    (2029, 2029),
                ],
                db_index=True,
                default=2017,
                verbose_name="Year",
            ),
        ),
    ]



# Generated by Django 2.0.6 on 2018-06-19 19:22

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import phonenumber_field.modelfields
import private_storage.fields
import private_storage.storage.files


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AnnualLeave',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('year',
                 models.PositiveIntegerField(
                     choices=[(2017, 2017), (2018, 2018), (2019, 2019),
                              (2020, 2020), (2021, 2021), (2022, 2022)],
                     db_index=True,
                     default=2018,
                     verbose_name='Year')),
                ('leave_type',
                 models.CharField(
                     choices=[('1', 'Sick Leave'), ('2', 'Regular Leave')],
                     db_index=True,
                     max_length=1,
                     verbose_name='Type')),
                ('allowed_days',
                 models.PositiveIntegerField(
                     blank=True,
                     default=21,
                     help_text='Number of leave days allowed in a year.',
                     verbose_name='Allowed Leave days')),
                ('carried_over_days',
                 models.PositiveIntegerField(
                     blank=True,
                     default=0,
                     help_text=
                     'Number of leave days carried over into this year.',
                     verbose_name='Carried Over Leave days')),
            ],
            options={
                'verbose_name': 'Annual Leave',
                'verbose_name_plural': 'Annual Leave',
                'ordering': ['year', 'leave_type', 'staff'],
            },
        ),
        migrations.CreateModel(
            name='Leave',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('start', models.DateTimeField(verbose_name='Start Date')),
                ('end', models.DateTimeField(verbose_name='End Date')),
                ('reason',
                 models.TextField(
                     blank=True, default='', verbose_name='Reason')),
                ('status',
                 models.CharField(
                     blank=True,
                     choices=[('1', 'Approved'), ('3', 'Pending'),
                              ('2', 'Rejected')],
                     db_index=True,
                     default='3',
                     max_length=1,
                     verbose_name='Status')),
                ('comments',
                 models.TextField(
                     blank=True, default='', verbose_name='Comments')),
                ('leave_type',
                 models.CharField(
                     blank=True,
                     choices=[('1', 'Sick Leave'), ('2', 'Regular Leave')],
                     db_index=True,
                     default='2',
                     max_length=1,
                     verbose_name='Type')),
            ],
            options={
                'verbose_name': 'Leave',
                'verbose_name_plural': 'Leave',
                'ordering': ['staff', 'start'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='OverTime',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('reason',
                 models.TextField(
                     blank=True, default='', verbose_name='Reason')),
                ('status',
                 models.CharField(
                     blank=True,
                     choices=[('1', 'Approved'), ('3', 'Pending'),
                              ('2', 'Rejected')],
                     db_index=True,
                     default='3',
                     max_length=1,
                     verbose_name='Status')),
                ('comments',
                 models.TextField(
                     blank=True, default='', verbose_name='Comments')),
                ('date', models.DateField(db_index=True, verbose_name='Date')),
                ('start', models.TimeField(verbose_name='Start')),
                ('end', models.TimeField(verbose_name='End')),
            ],
            options={
                'verbose_name': 'Overtime',
                'verbose_name_plural': 'Overtime',
                'ordering': ['staff', 'date', 'start'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('name', models.CharField(max_length=255,
                                          verbose_name='Name')),
                ('description',
                 models.TextField(
                     blank=True, default='', verbose_name='Description')),
            ],
            options={
                'verbose_name': 'Role',
                'verbose_name_plural': 'Roles',
                'ordering': ['name', 'created'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StaffDocument',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('name', models.CharField(max_length=255,
                                          verbose_name='Name')),
                ('description',
                 models.TextField(
                     blank=True, default='', verbose_name='Description')),
                ('file',
                 private_storage.fields.PrivateFileField(
                     help_text='Upload staff member drocument',
                     storage=private_storage.storage.files.
                     PrivateFileSystemStorage(),
                     upload_to='staff-documents/',
                     verbose_name='File')),
            ],
            options={
                'verbose_name': 'Staff Document',
                'verbose_name_plural': 'Staff Documents',
                'ordering': ['staff', 'name', 'created'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StaffProfile',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('created',
                 models.DateTimeField(
                     auto_now_add=True, verbose_name='Created')),
                ('modified',
                 models.DateTimeField(auto_now=True, verbose_name='Modified')),
                ('sex',
                 models.CharField(
                     blank=True,
                     choices=[('0', 'Not Known'), ('1', 'Male'),
                              ('2', 'Female'), ('9', 'Not Applicable')],
                     db_index=True,
                     default='0',
                     max_length=1,
                     verbose_name='Gender')),
                ('phone',
                 phonenumber_field.modelfields.PhoneNumberField(
                     blank=True,
                     default='',
                     max_length=128,
                     verbose_name='Phone')),
                ('address',
                 models.TextField(
                     blank=True, default='', verbose_name='Addresss')),
                ('birthday',
                 models.DateField(
                     blank=True,
                     default=None,
                     null=True,
                     verbose_name='Birth day')),
                ('leave_days',
                 models.PositiveIntegerField(
                     blank=True,
                     default=21,
                     help_text='Number of leave days allowed in a year.',
                     verbose_name='Leave days')),
                ('sick_days',
                 models.PositiveIntegerField(
                     blank=True,
                     default=10,
                     help_text='Number of sick days allowed in a year.',
                     verbose_name='Sick days')),
                ('overtime_allowed',
                 models.BooleanField(
                     default=False, verbose_name='Overtime allowed')),
                ('start_date',
                 models.DateField(
                     blank=True,
                     default=None,
                     help_text='The start date of employment',
                     null=True,
                     verbose_name='Start Date')),
                ('end_date',
                 models.DateField(
                     blank=True,
                     default=None,
                     help_text='The end date of employment',
                     null=True,
                     verbose_name='End Date')),
                ('data',
                 django.contrib.postgres.fields.jsonb.JSONField(
                     blank=True, default=dict, verbose_name='Data')),
                ('role',
                 models.ForeignKey(
                     blank=True,
                     default=None,
                     null=True,
                     on_delete=django.db.models.deletion.SET_NULL,
                     to='small_small_hr.Role',
                     verbose_name='Role')),
                ('user',
                 models.OneToOneField(
                     on_delete=django.db.models.deletion.CASCADE,
                     to=settings.AUTH_USER_MODEL,
                     verbose_name='User')),
            ],
            options={
                'verbose_name':
                'Staff Profile',
                'verbose_name_plural':
                'Staff Profiles',
                'ordering': [
                    'user__first_name', 'user__last_name', 'user__username',
                    'created'
                ],
                'abstract':
                False,
            },
        ),
        migrations.AddField(
            model_name='staffdocument',
            name='staff',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='small_small_hr.StaffProfile',
                verbose_name='Staff Member'),
        ),
        migrations.AddField(
            model_name='overtime',
            name='staff',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='small_small_hr.StaffProfile',
                verbose_name='Staff Member'),
        ),
        migrations.AddField(
            model_name='leave',
            name='staff',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='small_small_hr.StaffProfile',
                verbose_name='Staff Member'),
        ),
        migrations.AddField(
            model_name='annualleave',
            name='staff',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='small_small_hr.StaffProfile',
                verbose_name='Staff Member'),
        ),
        migrations.AlterUniqueTogether(
            name='annualleave',
            unique_together={('year', 'staff', 'leave_type')},
        ),
    ]



"""Main init file for small_small_hr."""
VERSION = (0, 4, 2)
__version__ = ".".join(str(v) for v in VERSION)
# pylint: disable=invalid-name
default_app_config = "small_small_hr.apps.SmallSmallHrConfig"  # noqa



# Generated by Django 2.1.7 on 2019-06-25 18:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("small_small_hr", "0006_auto_20181209_0108")]

    operations = [
        migrations.AlterField(
            model_name="annualleave",
            name="year",
            field=models.PositiveIntegerField(
                choices=[
                    (2017, 2017),
                    (2018, 2018),
                    (2019, 2019),
                    (2020, 2020),
                    (2021, 2021),
                    (2022, 2022),
                    (2023, 2023),
                    (2024, 2024),
                    (2025, 2025),
                    (2026, 2026),
                    (2027, 2027),
                    (2028, 2028),
                ],
                db_index=True,
                default=2017,
                verbose_name="Year",
            ),
        )
    ]



"""
Serializers for users app
"""
from django.contrib.auth.models import User  # pylint: disable = imported-auth-user

from rest_framework import serializers

from small_small_hr.models import StaffProfile


# pylint: disable=too-many-ancestors
class UserSerializer(serializers.ModelSerializer):
    """
    UserSerializer class
    """

    class Meta:  # pylint:  disable=too-few-public-methods
        """
        meta options
        """

        model = User
        fields = ("username", "first_name", "last_name", "email")


class StaffProfileSerializer(serializers.ModelSerializer):
    """
    Serializer class for StaffProfile model
    """

    first_name = serializers.CharField(source="user.first_name")
    last_name = serializers.CharField(source="user.last_name")
    id_number = serializers.SerializerMethodField()
    phone = serializers.SerializerMethodField()
    sex = serializers.SerializerMethodField()
    role = serializers.SerializerMethodField()
    nhif = serializers.SerializerMethodField()
    nssf = serializers.SerializerMethodField()
    pin_number = serializers.SerializerMethodField()
    address = serializers.SerializerMethodField()
    birthday = serializers.SerializerMethodField()
    leave_days = serializers.SerializerMethodField()
    sick_days = serializers.SerializerMethodField()
    overtime_allowed = serializers.SerializerMethodField()
    start_date = serializers.SerializerMethodField()
    end_date = serializers.SerializerMethodField()
    emergency_contact_name = serializers.SerializerMethodField()
    emergency_contact_number = serializers.SerializerMethodField()

    class Meta:  # pylint:  disable=too-few-public-methods
        """
        class meta options
        """

        model = StaffProfile
        fields = [
            "id",
            "first_name",
            "last_name",
            "created",
            "id_number",
            "phone",
            "sex",
            "modified",
            "role",
            "nhif",
            "emergency_contact_name",
            "nssf",
            "address",
            "birthday",
            "overtime_allowed",
            "leave_days",
            "start_date",
            "sick_days",
            "pin_number",
            "end_date",
            "emergency_contact_number",
        ]

    def get_id_number(self, obj):  # pylint: disable=no-self-use
        """
        Get id_number
        """
        return obj.data.get("id_number")

    def get_phone(self, obj):  # pylint: disable=no-self-use
        """
        Get phone
        """
        return obj.data.get("phone")

    def get_sex(self, obj):  # pylint: disable=no-self-use
        """
        Get sex
        """
        return obj.data.get("sex")

    def get_role(self, obj):  # pylint: disable=no-self-use
        """
        Get role
        """
        return obj.data.get("role")

    def get_nhif(self, obj):  # pylint: disable=no-self-use
        """
        Get nhhf
        """
        return obj.data.get("nhif")

    def get_nssf(self, obj):  # pylint: disable=no-self-use
        """
        Get nssf
        """
        return obj.data.get("nssf")

    def get_pin_number(self, obj):  # pylint: disable=no-self-use
        """
        Get pin_number
        """
        return obj.data.get("pin_number")

    def get_address(self, obj):  # pylint: disable=no-self-use
        """
        Get address
        """
        return obj.data.get("address")

    def get_birthday(self, obj):  # pylint: disable=no-self-use
        """
        Get birthday
        """
        return obj.data.get("birthday")

    def get_leave_days(self, obj):  # pylint: disable=no-self-use
        """
        Get leave_days
        """
        return obj.data.get("leave_days")

    def get_sick_days(self, obj):  # pylint: disable=no-self-use
        """
        Get sick_days
        """
        return obj.data.get("sick_days")

    def get_overtime_allowed(self, obj):  # pylint: disable=no-self-use
        """
        Get overtime_allowed
        """
        return obj.data.get("overtime_allowed")

    def get_start_date(self, obj):  # pylint: disable=no-self-use
        """
        Get start_date
        """
        return obj.data.get("start_date")

    def get_end_date(self, obj):  # pylint: disable=no-self-use
        """
        Get end_date
        """
        return obj.data.get("end_date")

    def get_emergency_contact_name(self, obj):  # pylint: disable=no-self-use
        """
        Get emergency_contact_name
        """
        return obj.data.get("emergency_contact_name")

    def get_emergency_contact_number(self, obj):  # pylint: disable=no-self-use
        """
        Get emergency_contact_number
        """
        return obj.data.get("emergency_contact_number")



# Generated by Django 3.1.2 on 2022-01-31 13:40

from decimal import Decimal
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0010_auto_20200626_1313'),
    ]

    operations = [
        migrations.AlterField(
            model_name='annualleave',
            name='allowed_days',
            field=models.DecimalField(blank=True, decimal_places=1, default=21, help_text='Number of leave days allowed in a year.', max_digits=12, validators=[django.core.validators.MinValueValidator(Decimal('0.1'))], verbose_name='Allowed Leave days'),
        ),
        migrations.AlterField(
            model_name='annualleave',
            name='carried_over_days',
            field=models.DecimalField(blank=True, decimal_places=1, default=0, help_text='Number of leave days carried over into this year.', max_digits=12, validators=[django.core.validators.MinValueValidator(Decimal('0.1'))], verbose_name='Carried Over Leave days'),
        ),
        migrations.AlterField(
            model_name='annualleave',
            name='year',
            field=models.PositiveIntegerField(choices=[(2017, 2017), (2018, 2018), (2019, 2019), (2020, 2020), (2021, 2021), (2022, 2022), (2023, 2023), (2024, 2024), (2025, 2025), (2026, 2026), (2027, 2027), (2028, 2028), (2029, 2029), (2030, 2030), (2031, 2031)], db_index=True, default=2017, verbose_name='Year'),
        ),
    ]



"""Review module for small-small-hr."""
from django.conf import settings
from django.contrib.auth.models import User  # pylint: disable = imported-auth-user
from django.db import models

from model_reviews.models import Reviewer

from small_small_hr.constants import STAFF


def set_staff_request_review_user(review_obj: models.Model):
    """
    Set user for Leave and Overtime requests.

    This is the default strategy of auto-setting the user for a review object.
    It simply sets the user using a field on the model object that is under review.
    """
    if not review_obj.user:
        object_under_review = review_obj.content_object
        staff_profile = getattr(object_under_review, STAFF, None)
        if staff_profile:
            review_obj.user = staff_profile.user


def set_staff_request_reviewer(review_obj: models.Model):
    """
    Set reviewer for Leave and Overtime requests.

    This is the strategy that will be used:

        1. Set it to the staff member's supervisor
        2. Additionally, set to all members of the Group named SSHR_ADMIN_USER_GROUP_NAME
    """
    if review_obj.user:
        staff_member = review_obj.user.staffprofile
        manager = staff_member.supervisor
        if (
            manager
            and not Reviewer.objects.filter(
                review=review_obj, user=manager.user
            ).exists()
        ):
            reviewer = Reviewer(review=review_obj, user=manager.user)
            reviewer.save()  # ensure save method is called

    hr_group_name = settings.SSHR_ADMIN_USER_GROUP_NAME
    for user in User.objects.filter(groups__name=hr_group_name):
        if not Reviewer.objects.filter(review=review_obj, user=user).exists():
            reviewer = Reviewer(review=review_obj, user=user)
            reviewer.save()  # ensure save method is called



"""
Small small HR signals module
"""
from django.conf import settings

from small_small_hr.models import StaffProfile

USER = settings.AUTH_USER_MODEL


# pylint: disable=unused-argument
def create_staffprofile(sender, instance, created, **kwargs):
    """
    Create staffprofile when a user object is created

    This signal is not connected by default
    """
    if created or not instance.staffprofile:
        # pylint: disable=unused-variable
        # pylint: disable=no-member
        profile, profile_created = \
            StaffProfile.objects.get_or_create(user=instance)



"""
Small small HR model managers module
"""
from django.db import models


class LeaveManager(models.Manager):  # pylint: disable=too-few-public-methods
    """
    Custom manager for Leave model
    """

    def get_queryset(self):
        """
        Get the queryset
        """
        return super().get_queryset().annotate(
            duration=models.F('end')-models.F('start'))



"""Constants."""
HOURS = "hours"
MINUTES = "minutes"
STAFF = "staff"
OVERTIME_APPLICATION_EMAIL_TEMPLATE = "overtime_application"
OVERTIME_COMPLETED_EMAIL_TEMPLATE = "overtime_completed"
LEAVE_APPLICATION_EMAIL_TEMPLATE = "leave_application"
LEAVE_COMPLETED_EMAIL_TEMPLATE = "leave_completed"
EMAIL_TEMPLATE_PATH = "small_small_hr/email"



# Generated by Django 2.1.3 on 2018-12-08 22:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0005_auto_20181115_2112'),
    ]

    operations = [
        migrations.CreateModel(
            name='FreeDay',
            fields=[
                ('id',
                 models.AutoField(
                     auto_created=True,
                     primary_key=True,
                     serialize=False,
                     verbose_name='ID')),
                ('name', models.CharField(max_length=255,
                                          verbose_name='Name')),
                ('date', models.DateField(unique=True, verbose_name='Date')),
            ],
            options={
                'verbose_name': 'Free Day',
                'verbose_name_plural': 'Free Days',
                'ordering': ['-date'],
            },
        ),
        migrations.AlterModelOptions(
            name='annualleave',
            options={
                'ordering': ['-year', 'leave_type', 'staff'],
                'verbose_name': 'Annual Leave',
                'verbose_name_plural': 'Annual Leave'
            },
        ),
        migrations.AlterModelOptions(
            name='leave',
            options={
                'ordering': ['staff', '-start'],
                'verbose_name': 'Leave',
                'verbose_name_plural': 'Leave'
            },
        ),
        migrations.AlterModelOptions(
            name='overtime',
            options={
                'ordering': ['staff', '-date', 'start'],
                'verbose_name': 'Overtime',
                'verbose_name_plural': 'Overtime'
            },
        ),
        migrations.AlterModelOptions(
            name='staffdocument',
            options={
                'ordering': ['staff', 'name', '-created'],
                'verbose_name': 'Staff Document',
                'verbose_name_plural': 'Staff Documents'
            },
        ),
        migrations.AlterField(
            model_name='annualleave',
            name='year',
            field=models.PositiveIntegerField(
                choices=[(2017, 2017), (2018, 2018), (2019, 2019), (2020,
                                                                    2020),
                         (2021, 2021), (2022, 2022), (2023, 2023), (2024,
                                                                    2024),
                         (2025, 2025), (2026, 2026), (2027, 2027)],
                db_index=True,
                default=2018,
                verbose_name='Year'),
        ),
    ]



"""
Utils module for small small hr
"""
from datetime import date

from django.conf import settings
from django.utils import timezone

from small_small_hr.models import AnnualLeave, FreeDay, Leave, StaffProfile


def get_carry_over(staffprofile: StaffProfile, year: int, leave_type: str):
    """Get carried over leave days."""
    # pylint: disable=no-member
    if leave_type == Leave.REGULAR:
        previous_obj = AnnualLeave.objects.filter(
            staff=staffprofile, year=year - 1, leave_type=leave_type
        ).first()
        if previous_obj:
            remaining = previous_obj.get_available_leave_days()
            max_carry_over = settings.SSHR_MAX_CARRY_OVER
            if remaining > max_carry_over:
                carry_over = max_carry_over
            else:
                carry_over = remaining

            return carry_over

    return 0


def create_annual_leave(staffprofile: StaffProfile, year: int, leave_type: str):
    """Creates an annual leave object for the staff member."""
    # pylint: disable=no-member
    try:
        annual_leave = AnnualLeave.objects.get(
            staff=staffprofile, year=year, leave_type=leave_type
        )
    except AnnualLeave.DoesNotExist:
        carry_over = get_carry_over(staffprofile, year, leave_type)

        if leave_type == Leave.REGULAR:
            allowed_days = staffprofile.leave_days
        elif leave_type == Leave.SICK:
            allowed_days = staffprofile.sick_days

        annual_leave = AnnualLeave(
            staff=staffprofile,
            year=year,
            leave_type=leave_type,
            allowed_days=allowed_days,
            carried_over_days=carry_over,
        )
        annual_leave.save()

    return annual_leave


def create_free_days(start_year: int = timezone.now().year, number_of_years: int = 11):
    """
    Create FreeDay records.

    :param start_year:  the year from which to start creating free days
    :param number_of_years: number of years to create free days objects
    """
    default_days = settings.SSHR_FREE_DAYS
    years = (start_year + _ for _ in range(number_of_years))
    for year in years:
        for default_day in default_days:
            the_date = date(
                year=year, month=default_day["month"], day=default_day["day"],
            )
            free_day = FreeDay(name=the_date.strftime("%A %d %B %Y"), date=the_date,)
            free_day.save()



"""Emails module for scam app."""
from model_reviews.emails import get_display_name, send_email
from model_reviews.models import ModelReview, Reviewer

from small_small_hr.constants import (
    LEAVE_APPLICATION_EMAIL_TEMPLATE,
    LEAVE_COMPLETED_EMAIL_TEMPLATE,
    OVERTIME_APPLICATION_EMAIL_TEMPLATE,
    OVERTIME_COMPLETED_EMAIL_TEMPLATE,
)


def send_request_for_leave_review(reviewer: Reviewer):
    """Send email requesting a Leave review to one reviewer."""
    if reviewer.user.email:
        source = reviewer.review.content_object
        send_email(
            name=get_display_name(reviewer.user),
            email=reviewer.user.email,
            subject=source.review_request_email_subject,
            message=source.review_request_email_body,
            obj=reviewer.review,
            cc_list=None,
            template=LEAVE_APPLICATION_EMAIL_TEMPLATE,
            template_path=source.email_template_path,
        )


def send_request_for_overtime_review(reviewer: Reviewer):
    """Send email requesting a OverTime review to one reviewer."""
    if reviewer.user.email:
        source = reviewer.review.content_object
        send_email(
            name=get_display_name(reviewer.user),
            email=reviewer.user.email,
            subject=source.review_request_email_subject,
            message=source.review_request_email_body,
            obj=reviewer.review,
            cc_list=None,
            template=OVERTIME_APPLICATION_EMAIL_TEMPLATE,
            template_path=source.email_template_path,
        )


def send_leave_review_complete_notice(review_obj: ModelReview):
    """Send notice that Leave review is complete."""
    if not review_obj.needs_review() and review_obj.user:
        if review_obj.user.email:
            source = review_obj.content_object
            send_email(
                name=get_display_name(review_obj.user),
                email=review_obj.user.email,
                subject=source.review_complete_email_subject,
                message=source.review_complete_email_body,
                obj=review_obj,
                cc_list=None,
                template=LEAVE_COMPLETED_EMAIL_TEMPLATE,
                template_path=source.email_template_path,
            )


def send_overtime_review_complete_notice(review_obj: ModelReview):
    """Send notice that OverTime review is complete."""
    if not review_obj.needs_review() and review_obj.user:
        if review_obj.user.email:
            source = review_obj.content_object
            send_email(
                name=get_display_name(review_obj.user),
                email=review_obj.user.email,
                subject=source.review_complete_email_subject,
                message=source.review_complete_email_body,
                obj=review_obj,
                cc_list=None,
                template=OVERTIME_COMPLETED_EMAIL_TEMPLATE,
                template_path=source.email_template_path,
            )



"""Setup.py for small_small_hr."""
import os
import sys

from setuptools import find_packages, setup

import small_small_hr as sshr

# read the contents of your README file
with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as f:
    LONG_DESCRIPTION = f.read()

# convenience command to publish
if sys.argv[-1] == "publish":
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("rm -rf build/ *.egg-info/")
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/* --skip-existing")
    print("You probably want to also tag the version now:")
    print(f"  git tag -a v{sshr.__version__} -m 'version {sshr.__version__}'")
    print("  git push --tags")
    sys.exit()

setup(
    name="small-small-hr",
    version=sshr.__version__,
    description="Minimal human resource management app for Django",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Kelvin Jayanoris",
    author_email="kelvin@jayanoris.com",
    url="https://github.com/moshthepitt/small-small-hr",
    packages=find_packages(exclude=["docs", "*.egg-info", "build", "tests.*", "tests"]),
    install_requires=[
        "Django >= 2.2",
        "voluptuous",
        "psycopg2-binary",
        "sorl-thumbnail",
        "django-private-storage",
        "django-model-reviews",
        "phonenumberslite",
        "django-phonenumber-field",
        "django-crispy-forms",
        "djangorestframework",
        "sorl-thumbnail",
        "Pillow",
        "django-mptt",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
    ],
    include_package_data=True,
)



# Generated by Django 2.1.2 on 2018-11-15 18:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('small_small_hr', '0004_auto_20180725_2127'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffdocument',
            name='public',
            field=models.BooleanField(
                blank=True,
                default=False,
                help_text='If public, it will be available to everyone.',
                verbose_name='Public'),
        ),
        migrations.AlterField(
            model_name='staffprofile',
            name='overtime_allowed',
            field=models.BooleanField(
                blank=True, default=False, verbose_name='Overtime allowed'),
        ),
    ]



"""Forms module for small small hr."""
from datetime import datetime, time

from django import forms
from django.conf import settings
from django.contrib.auth.models import User  # pylint: disable = imported-auth-user
from django.db.models import Q
from django.utils.translation import ugettext as _

import pytz
from crispy_forms.bootstrap import Field, FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit
from phonenumber_field.formfields import PhoneNumberField
from phonenumber_field.phonenumber import PhoneNumber

from small_small_hr.models import (
    TWOPLACES,
    AnnualLeave,
    FreeDay,
    Leave,
    OverTime,
    Role,
    StaffDocument,
    StaffProfile,
)


class AnnualLeaveForm(forms.ModelForm):
    """Form used when managing AnnualLeave."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = AnnualLeave
        fields = ["staff", "year", "leave_type", "allowed_days", "carried_over_days"]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        if not self.instance:
            self.fields["year"].initial = datetime.today().year
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "annual-leave-form"
        self.helper.layout = Layout(
            Field("staff",),
            Field("year",),
            Field("leave_type",),
            Field("allowed_days",),
            Field("carried_over_days"),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class RoleForm(forms.ModelForm):
    """Form used when managing Role objects."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = Role
        fields = ["name", "description"]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "role-form"
        self.helper.layout = Layout(
            Field("name",),
            Field("description",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class FreeDayForm(forms.ModelForm):
    """Form used when managing FreeDay objects."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = FreeDay
        fields = ["name", "date"]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "freeday-form"
        self.helper.layout = Layout(
            Field("name",),
            Field("date",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class OverTimeForm(forms.ModelForm):
    """Form used when managing OverTime objects."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = OverTime
        fields = [
            "staff",
            "date",
            "start",
            "end",
            "review_reason",
            "review_status",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "overtime-form"
        self.helper.layout = Layout(
            Field("staff",),
            Field("date",),
            Field("start",),
            Field("end",),
            Field("review_reason",),
            Field("review_status",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )

    def clean(self):
        """Clean all the form fields."""
        cleaned_data = super().clean()
        end = cleaned_data.get("end")
        start = cleaned_data.get("start")
        date = cleaned_data.get("date")
        staff = cleaned_data.get("staff")
        review_status = cleaned_data.get("review_status")

        # end must be later than start
        if end <= start:
            self.add_error("end", _("end must be greater than start"))

        # must not overlap within the same date unless being rejected
        # pylint: disable=no-member
        overlap_qs = OverTime.objects.filter(
            date=date, staff=staff, review_status=OverTime.APPROVED
        ).filter(Q(start__gte=start) & Q(end__lte=end))

        if self.instance is not None:
            overlap_qs = overlap_qs.exclude(id=self.instance.id)

        if overlap_qs.exists() and review_status != OverTime.REJECTED:
            msg = _("you cannot have overlapping overtime hours on the " "same day")
            self.add_error("start", msg)
            self.add_error("end", msg)
            self.add_error("date", msg)


class ApplyOverTimeForm(OverTimeForm):
    """Form used when applying for overtime."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = OverTime
        fields = [
            "staff",
            "date",
            "start",
            "end",
            "review_reason",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        super().__init__(*args, **kwargs)
        self.request = kwargs.pop("request", None)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields["staff"].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id
                )
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "overtime-application-form"
        self.helper.layout = Layout(
            Field("staff", type="hidden"),
            Field("date",),
            Field("start",),
            Field("end",),
            Field("review_reason",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )

    def save(self, commit=True):
        """Save the form."""
        overtime = super().save()
        return overtime


class LeaveForm(forms.ModelForm):
    """Form used when managing Leave objects."""

    start = forms.DateField(label=_("Start Date"), required=True)
    end = forms.DateField(label=_("End Date"), required=True)

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = Leave
        fields = [
            "staff",
            "leave_type",
            "start",
            "end",
            "review_reason",
            "review_status",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "leave-form"
        self.helper.layout = Layout(
            Field("staff",),
            Field("leave_type",),
            Field("start",),
            Field("end",),
            Field("review_reason",),
            Field("review_status",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )

    def clean_start(self):
        """Clean start field."""
        data = self.cleaned_data["start"]
        data = datetime.combine(
            date=data,
            time=time(settings.SSHR_DEFAULT_TIME, 0, 0, 0),
            tzinfo=pytz.timezone(settings.TIME_ZONE),
        )
        return data

    def clean_end(self):
        """Clean end field."""
        data = self.cleaned_data["end"]
        data = datetime.combine(
            date=data,
            time=time(settings.SSHR_DEFAULT_TIME, 0, 0, 0),
            tzinfo=pytz.timezone(settings.TIME_ZONE),
        )
        return data

    def clean(self):
        """Clean all the form fields."""
        cleaned_data = super().clean()
        leave_type = cleaned_data.get("leave_type")
        staff = cleaned_data.get("staff")
        end = cleaned_data.get("end")
        start = cleaned_data.get("start")
        review_status = cleaned_data.get("review_status")

        if all([staff, leave_type, start, end]):
            # end year and start year must be the same
            if end.year != start.year:
                msg = _("start and end must be from the same year")
                self.add_error("start", msg)
                self.add_error("end", msg)

            # end must be later than start
            if end < start:
                self.add_error("end", _("end must be greater than start"))

            if not settings.SSHR_ALLOW_OVERSUBSCRIBE:
                # staff profile must have sufficient sick days
                if leave_type == Leave.SICK:
                    sick_days = staff.get_available_sick_days(year=start.year)
                    if (end - start).days > sick_days:
                        msg = _(
                            "Not enough sick days. Available sick days "
                            f"are {sick_days.quantize(TWOPLACES)}"
                        )
                        self.add_error("start", msg)
                        self.add_error("end", msg)

                # staff profile must have sufficient leave days
                if leave_type == Leave.REGULAR:
                    leave_days = staff.get_available_leave_days(year=start.year)
                    if (end - start).days > leave_days:
                        msg = _(
                            "Not enough leave days. Available leave days "
                            f"are {leave_days.quantize(TWOPLACES)}"
                        )
                        self.add_error("start", msg)
                        self.add_error("end", msg)

            # must not overlap unless it is being rejected
            # pylint: disable=no-member
            overlap_qs = Leave.objects.filter(
                staff=staff, review_status=Leave.APPROVED, leave_type=leave_type
            ).filter(Q(start__gte=start) & Q(end__lte=end))

            if self.instance is not None:
                overlap_qs = overlap_qs.exclude(id=self.instance.id)

            if overlap_qs.exists() and review_status != Leave.REJECTED:
                msg = _("you cannot have overlapping leave days")
                self.add_error("start", msg)
                self.add_error("end", msg)


class ApplyLeaveForm(LeaveForm):
    """Form used when applying for Leave."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = Leave
        fields = [
            "staff",
            "leave_type",
            "start",
            "end",
            "review_reason",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        super().__init__(*args, **kwargs)
        self.request = kwargs.pop("request", None)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields["staff"].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id
                )
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "leave-application-form"
        self.helper.layout = Layout(
            Field("staff", type="hidden"),
            Field("leave_type",),
            Field("start",),
            Field("end",),
            Field("review_reason",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )

    def save(self, commit=True):
        """Save the form."""
        leave = super().save()
        return leave


class StaffDocumentForm(forms.ModelForm):
    """Form used when managing StaffDocument objects."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = StaffDocument
        fields = [
            "staff",
            "name",
            "description",
            "public",
            "file",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.file:
            self.fields["file"].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "staffdocument-form"
        self.helper.layout = Layout(
            Field("staff",),
            Field("name",),
            Field("description",),
            Field("file",),
            Field("public",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class UserStaffDocumentForm(forms.ModelForm):
    """Form used when managing one's own StaffDocument objects."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = StaffDocument
        fields = [
            "staff",
            "name",
            "description",
            "file",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        if self.request:
            # pylint: disable=no-member
            try:
                self.request.user.staffprofile
            except StaffProfile.DoesNotExist:
                pass
            else:
                self.fields["staff"].queryset = StaffProfile.objects.filter(
                    id=self.request.user.staffprofile.id
                )
        if self.instance and self.instance.file:
            self.fields["file"].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "staffdocument-form"
        self.helper.layout = Layout(
            Field("staff", type="hidden"),
            Field("name",),
            Field("description",),
            Field("file",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class StaffProfileAdminForm(forms.ModelForm):
    """Form used when managing StaffProfile objects."""

    first_name = forms.CharField(label=_("First Name"), required=True)
    last_name = forms.CharField(label=_("Last Name"), required=True)
    id_number = forms.CharField(label=_("ID Number"), required=True)
    nhif = forms.CharField(label=_("NHIF"), required=False)
    nssf = forms.CharField(label=_("NSSF"), required=False)
    pin_number = forms.CharField(label=_("PIN Number"), required=False)
    emergency_contact_name = forms.CharField(
        label=_("Emergency Contact Name"), required=False
    )
    emergency_contact_relationship = forms.CharField(
        label=_("Emergency Contact Relationship"), required=False
    )
    emergency_contact_number = PhoneNumberField(
        label=_("Emergency Contact Phone Number"), required=False
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = StaffProfile
        fields = [
            "first_name",
            "last_name",
            "id_number",
            "image",
            "phone",
            "sex",
            "role",
            "nhif",
            "nssf",
            "pin_number",
            "address",
            "birthday",
            "leave_days",
            "sick_days",
            "overtime_allowed",
            "start_date",
            "end_date",
            "emergency_contact_name",
            "emergency_contact_number",
            "emergency_contact_relationship",
            "supervisor",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.image:
            self.fields["image"].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "staffprofile-form"
        self.helper.layout = Layout(
            Field("first_name",),
            Field("last_name",),
            Field("supervisor",),
            Field("image",),
            Field("phone",),
            Field("id_number",),
            Field("sex",),
            Field("role",),
            Field("nhif",),
            Field("nssf",),
            Field("pin_number",),
            Field("address",),
            Field("birthday",),
            Field("leave_days",),
            Field("sick_days",),
            Field("overtime_allowed",),
            Field("start_date",),
            Field("end_date",),
            Field("emergency_contact_name",),
            Field("emergency_contact_number",),
            Field("emergency_contact_relationship",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )

    def clean_id_number(self):
        """Check if id number is unique."""
        value = self.cleaned_data.get("id_number")
        # pylint: disable=no-member
        if (
            StaffProfile.objects.exclude(id=self.instance.id)
            .filter(data__id_number=value)
            .exists()
        ):
            raise forms.ValidationError(_("This id number is already in use."))
        return value

    def clean_nssf(self):
        """Check if NSSF number is unique."""
        value = self.cleaned_data.get("nssf")
        # pylint: disable=no-member
        if (
            value
            and StaffProfile.objects.exclude(id=self.instance.id)
            .filter(data__nssf=value)
            .exists()
        ):
            raise forms.ValidationError(_("This NSSF number is already in use."))
        return value

    def clean_nhif(self):
        """Check if NHIF number is unique."""
        value = self.cleaned_data.get("nhif")
        # pylint: disable=no-member
        if (
            value
            and StaffProfile.objects.exclude(id=self.instance.id)
            .filter(data__nhif=value)
            .exists()
        ):
            raise forms.ValidationError(_("This NHIF number is already in use."))
        return value

    def clean_pin_number(self):
        """Check if PIN number is unique."""
        value = self.cleaned_data.get("pin_number")
        # pylint: disable=no-member
        if (
            value
            and StaffProfile.objects.exclude(id=self.instance.id)
            .filter(data__pin_number=value)
            .exists()
        ):
            raise forms.ValidationError(_("This PIN number is already in use."))
        return value

    def save(self, commit=True):  # pylint: disable=unused-argument
        """Save the form."""
        staffprofile = super().save()

        emergency_phone = self.cleaned_data.get("emergency_contact_number")
        if isinstance(emergency_phone, PhoneNumber):
            emergency_phone = emergency_phone.as_e164

        json_data = {
            "id_number": self.cleaned_data.get("id_number"),
            "nhif": self.cleaned_data.get("nhif"),
            "nssf": self.cleaned_data.get("nssf"),
            "pin_number": self.cleaned_data.get("pin_number"),
            "emergency_contact_name": self.cleaned_data.get("emergency_contact_name"),
            "emergency_contact_relationship": self.cleaned_data.get(
                "emergency_contact_relationship"
            ),
            "emergency_contact_number": emergency_phone,
        }
        staffprofile.data = json_data
        staffprofile.save()

        user = staffprofile.user
        user.first_name = self.cleaned_data["first_name"]
        user.last_name = self.cleaned_data["last_name"]
        user.save()
        return staffprofile


class StaffProfileAdminCreateForm(StaffProfileAdminForm):
    """Form used when creating new Staff Profiles."""

    user = forms.ModelChoiceField(
        label=_("User"), queryset=User.objects.filter(staffprofile=None)
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = StaffProfile
        fields = [
            "user",
            "first_name",
            "last_name",
            "supervisor",
            "id_number",
            "image",
            "phone",
            "sex",
            "role",
            "nhif",
            "nssf",
            "pin_number",
            "address",
            "birthday",
            "leave_days",
            "sick_days",
            "overtime_allowed",
            "start_date",
            "end_date",
            "emergency_contact_name",
            "emergency_contact_number",
            "emergency_contact_relationship",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.image:
            self.fields["image"].required = False
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "staffprofile-form"
        self.helper.layout = Layout(
            Field("user",),
            Field("first_name",),
            Field("last_name",),
            Field("supervisor",),
            Field("image",),
            Field("phone",),
            Field("id_number",),
            Field("sex",),
            Field("role",),
            Field("nhif",),
            Field("nssf",),
            Field("pin_number",),
            Field("address",),
            Field("birthday",),
            Field("leave_days",),
            Field("sick_days",),
            Field("overtime_allowed",),
            Field("start_date",),
            Field("end_date",),
            Field("emergency_contact_name",),
            Field("emergency_contact_number",),
            Field("emergency_contact_relationship",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )


class StaffProfileUserForm(StaffProfileAdminForm):
    """Form used when the user is updating their own data."""

    class Meta:  # pylint: disable=too-few-public-methods
        """Class meta options."""

        model = StaffProfile
        fields = [
            "first_name",
            "last_name",
            "id_number",
            "image",
            "phone",
            "sex",
            "nhif",
            "nssf",
            "pin_number",
            "address",
            "birthday",
            "emergency_contact_name",
            "emergency_contact_number",
            "emergency_contact_relationship",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = "post"
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.form_id = "staffprofile-user-form"
        self.helper.layout = Layout(
            Field("first_name",),
            Field("last_name",),
            Field("image",),
            Field("phone",),
            Field("id_number",),
            Field("sex",),
            Field("nhif",),
            Field("nssf",),
            Field("pin_number",),
            Field("address",),
            Field("birthday",),
            Field("emergency_contact_name",),
            Field("emergency_contact_number",),
            Field("emergency_contact_relationship",),
            FormActions(Submit("submitBtn", _("Submit"), css_class="btn-primary"),),
        )



"""
Configurable options
"""
from django.conf import settings

SSHR_MAX_CARRY_OVER = 10
SSHR_DAY_LEAVE_VALUES = {
    1: 1,  # Monday
    2: 1,  # Tuesday
    3: 1,  # Wednesday
    4: 1,  # Thursday
    5: 1,  # Friday
    6: 0,  # Saturday
    7: 0,  # Sunday
}
SSHR_ALLOW_OVERSUBSCRIBE = True  # allow taking more leave days one has
SSHR_DEFAULT_TIME = 7  # default time of the day for leave
SSHR_FREE_DAYS = [
    {"day": 1, "month": 1},  # New year
    {"day": 1, "month": 5},  # labour day
    {"day": 1, "month": 6},  # Madaraka day
    {"day": 20, "month": 10},  # Mashujaa day
    {"day": 12, "month": 12},  # Jamhuri day
    {"day": 25, "month": 12},  # Christmas
    {"day": 26, "month": 12},  # Boxing day
]  # these are days that are not counted when getting taken leave days
# admins
SSHR_ADMIN_USER_GROUP_NAME = "Human Resource"
# emails
SSHR_ADMIN_NAME = "HR"
SSHR_ADMIN_EMAILS = [settings.DEFAULT_FROM_EMAIL]
SSHR_ADMIN_LEAVE_EMAILS = SSHR_ADMIN_EMAILS
SSHR_ADMIN_OVERTIME_EMAILS = SSHR_ADMIN_EMAILS
# SSHR_LEAVE_PROCESSED_EMAIL_TXT - text of processed leave email
# SSHR_LEAVE_PROCESSED_EMAIL_SUBJ - subject of processed leave email
# SSHR_LEAVE_APPLICATION_EMAIL_TXT - text of leave application email
# SSHR_LEAVE_APPLICATION_EMAIL_SUBJ - subject of leave application email
# SSHR_OVERTIME_PROCESSED_EMAIL_TXT - text of processed overtime email
# SSHR_OVERTIME_PROCESSED_EMAIL_SUBJ - subject of processed overtime email
# SSHR_OVERTIME_APPLICATION_EMAIL_TXT - text of overtime application email
# SSHR_OVERTIME_APPLICATION_EMAIL_SUBJ - subject of overtime application email



"""Models module for small_small_hr."""
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Optional

from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.db import models
from django.core.validators import MinValueValidator
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _

from model_reviews.models import AbstractReview
from mptt.models import MPTTModel, TreeForeignKey
from phonenumber_field.modelfields import PhoneNumberField
from private_storage.fields import PrivateFileField
from sorl.thumbnail import ImageField

from small_small_hr.constants import EMAIL_TEMPLATE_PATH
from small_small_hr.managers import LeaveManager

USER = settings.AUTH_USER_MODEL
TWOPLACES = Decimal(10) ** -2


class TimeStampedModel(models.Model):
    """Abstract model class that includes timestamp fields."""

    created = models.DateTimeField(verbose_name=_("Created"), auto_now_add=True)
    modified = models.DateTimeField(verbose_name=_("Modified"), auto_now=True)

    # pylint: disable=too-few-public-methods
    class Meta:
        """Meta options for TimeStampedModel."""

        abstract = True


class Role(TimeStampedModel, models.Model):
    """Model class for staff member role."""

    name = models.CharField(_("Name"), max_length=255)
    description = models.TextField(_("Description"), blank=True, default="")

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for StaffDocument."""

        abstract = False
        verbose_name = _("Role")
        verbose_name_plural = _("Roles")
        ordering = ["name", "created"]

    def __str__(self):
        """Unicode representation of class object."""
        # pylint: disable=no-member
        return self.name


class StaffProfile(TimeStampedModel, MPTTModel):
    """
    StaffProfile model class.

    Extends auth.User and adds more fields
    """

    # sex choices
    # according to https://en.wikipedia.org/wiki/ISO/IEC_5218
    NOT_KNOWN = "0"
    MALE = "1"
    FEMALE = "2"
    NOT_APPLICABLE = "9"

    SEX_CHOICES = (
        (NOT_KNOWN, _("Not Known")),
        (MALE, _("Male")),
        (FEMALE, _("Female")),
        (NOT_APPLICABLE, _("Not Applicable")),
    )

    user = models.OneToOneField(USER, verbose_name=_("User"), on_delete=models.CASCADE)
    supervisor = TreeForeignKey(
        "self",
        verbose_name=_("Manager"),
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="children",
    )
    image = ImageField(
        upload_to="staff-images/",
        max_length=255,
        verbose_name=_("Profile Image"),
        help_text=_("A square image works best"),
        blank=True,
    )
    sex = models.CharField(
        _("Gender"),
        choices=SEX_CHOICES,
        max_length=1,
        default=NOT_KNOWN,
        blank=True,
        db_index=True,
    )
    role = models.ForeignKey(
        Role,
        verbose_name=_("Role"),
        blank=True,
        default=None,
        null=True,
        on_delete=models.SET_NULL,
    )
    phone = PhoneNumberField(_("Phone"), blank=True, default="")
    address = models.TextField(_("Addresss"), blank=True, default="")
    birthday = models.DateField(_("Birthday"), blank=True, default=None, null=True)
    leave_days = models.PositiveIntegerField(
        _("Leave days"),
        default=21,
        blank=True,
        help_text=_("Number of leave days allowed in a year."),
    )
    sick_days = models.PositiveIntegerField(
        _("Sick days"),
        default=10,
        blank=True,
        help_text=_("Number of sick days allowed in a year."),
    )
    overtime_allowed = models.BooleanField(
        _("Overtime allowed"), blank=True, default=False
    )
    start_date = models.DateField(
        _("Start Date"),
        null=True,
        default=None,
        blank=True,
        help_text=_("The start date of employment"),
    )
    end_date = models.DateField(
        _("End Date"),
        null=True,
        default=None,
        blank=True,
        help_text=_("The end date of employment"),
    )
    data = JSONField(_("Data"), default=dict, blank=True)

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for StaffProfile."""

        abstract = False
        verbose_name = _("Staff Profile")
        verbose_name_plural = _("Staff Profiles")
        ordering = ["user__first_name", "user__last_name", "user__username", "created"]

    class MPTTMeta:
        """Meta options for MPTT."""

        parent_attr = "supervisor"

    def get_name(self):
        """Return the staff member's name."""
        # pylint: disable=no-member
        return f"{self.user.first_name} {self.user.last_name}"

    @cached_property
    def _current_year(self):  # pylint: disable=no-self-use
        """Get the current year."""
        return datetime.today().year

    def get_approved_leave_days(self, year: Optional[int] = None):
        """Get approved leave days in the current year."""
        # pylint: disable=no-member
        return get_taken_leave_days(
            staffprofile=self,
            status=Leave.APPROVED,
            leave_type=Leave.REGULAR,
            start_year=year or self._current_year,
            end_year=year or self._current_year,
        )

    def get_approved_sick_days(self, year: Optional[int] = None):
        """Get approved leave days in the current year."""
        return get_taken_leave_days(
            staffprofile=self,
            status=Leave.APPROVED,
            leave_type=Leave.SICK,
            start_year=year or self._current_year,
            end_year=year or self._current_year,
        )

    def get_available_leave_days(self, year: Optional[int] = None):
        """Get available leave days."""
        try:
            # pylint: disable=no-member
            leave_record = AnnualLeave.objects.get(
                leave_type=Leave.REGULAR, staff=self, year=year or self._current_year
            )
        except AnnualLeave.DoesNotExist:
            return Decimal(0)
        else:
            return leave_record.get_available_leave_days()

    def get_available_sick_days(self, year: Optional[int] = None):
        """Get available sick days."""
        try:
            # pylint: disable=no-member
            leave_record = AnnualLeave.objects.get(
                leave_type=Leave.SICK, staff=self, year=year or self._current_year
            )
        except AnnualLeave.DoesNotExist:
            return Decimal(0)
        else:
            return leave_record.get_available_leave_days()

    def __str__(self):
        """Unicode representation of class object."""
        return self.get_name()  # pylint: disable=no-member


class StaffDocument(TimeStampedModel, models.Model):
    """StaffDocument model class."""

    staff = models.ForeignKey(
        StaffProfile, verbose_name=_("Staff Member"), on_delete=models.CASCADE
    )
    name = models.CharField(_("Name"), max_length=255)
    description = models.TextField(_("Description"), blank=True, default="")
    file = PrivateFileField(
        _("File"),
        upload_to="staff-documents/",
        help_text=_("Upload staff member document"),
        content_types=[
            "application/pdf",
            "application/msword",
            "application/vnd.oasis.opendocument.text",
            "image/jpeg",
            "image/png",
        ],
        max_file_size=10485760,
    )
    public = models.BooleanField(
        _("Public"),
        help_text=_("If public, it will be available to everyone."),
        blank=True,
        default=False,
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for StaffDocument."""

        abstract = False
        verbose_name = _("Staff Document")
        verbose_name_plural = _("Staff Documents")
        ordering = ["staff", "name", "-created"]

    def __str__(self):
        """Unicode representation of class object."""
        return f"{self.staff.get_name()} - {self.name}"


class BaseStaffRequest(TimeStampedModel, AbstractReview):
    """Abstract model class for Leave & Overtime tracking."""

    staff = models.ForeignKey(
        StaffProfile, verbose_name=_("Staff Member"), on_delete=models.CASCADE
    )
    start = models.DateTimeField(_("Start Date"))
    end = models.DateTimeField(_("End Date"))
    review_reason = models.TextField(_("Reason"), blank=True, default="")
    review_status = models.CharField(
        _("Status"),
        max_length=1,
        choices=AbstractReview.STATUS_CHOICES,
        default=AbstractReview.PENDING,
        blank=True,
        db_index=True,
    )

    # MODEL REVIEW OPTIONS
    # path to function that will be used to determine reviewers
    set_reviewers_function: Optional[
        str
    ] = "small_small_hr.reviews.set_staff_request_reviewer"
    # path to function that will be used to determine the user for a review object
    set_user_function: Optional[
        str
    ] = "small_small_hr.reviews.set_staff_request_review_user"

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for BaseStaffRequest."""

        abstract = True


class Leave(BaseStaffRequest):
    """Leave model class."""

    SICK = "1"
    REGULAR = "2"

    TYPE_CHOICES = (
        (SICK, _("Sick Leave")),
        (REGULAR, _("Regular Leave")),
    )

    leave_type = models.CharField(
        _("Type"),
        max_length=1,
        choices=TYPE_CHOICES,
        default=REGULAR,
        blank=True,
        db_index=True,
    )

    objects = LeaveManager()

    # MODEL REVIEW OPTIONS
    email_template_path = EMAIL_TEMPLATE_PATH
    # path to function that will be used to send email to reviewers
    request_for_review_function: Optional[
        str
    ] = "small_small_hr.emails.send_request_for_leave_review"
    # path to function that will be used to send email to user after review
    review_complete_notify_function: Optional[
        str
    ] = "small_small_hr.emails.send_leave_review_complete_notice"

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for Leave."""

        abstract = False
        verbose_name = _("Leave")
        verbose_name_plural = _("Leave")
        ordering = ["staff", "-start"]

    def __str__(self):
        """Unicode representation of class object."""
        # pylint: disable=no-member
        return _(f"{self.staff.get_name()}: {self.start} to {self.end}")

    def get_duration(self):
        """Get duration."""
        return self.end - self.start

    @cached_property
    def duration(self):
        """Get duration as a property."""
        return self.get_duration()

    @cached_property
    def day_count(self):
        """
        Get the number of leave days as a property.

        This takes into account holidays and weekend policy.
        """
        return get_real_leave_duration(leave_obj=self)


class OverTime(BaseStaffRequest):
    """Overtime model class."""

    date = models.DateField(
        _("Date"), auto_now=False, auto_now_add=False, db_index=True
    )
    start = models.TimeField(_("Start"), auto_now=False, auto_now_add=False)
    end = models.TimeField(_("End"), auto_now=False, auto_now_add=False)

    # MODEL REVIEW OPTIONS
    email_template_path = EMAIL_TEMPLATE_PATH
    # path to function that will be used to send email to reviewers
    request_for_review_function: Optional[
        str
    ] = "small_small_hr.emails.send_request_for_overtime_review"
    # path to function that will be used to send email to user after review
    review_complete_notify_function: Optional[
        str
    ] = "small_small_hr.emails.send_overtime_review_complete_notice"

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for OverTime."""

        abstract = False
        verbose_name = _("Overtime")
        verbose_name_plural = _("Overtime")
        ordering = ["staff", "-date", "start"]

    def __str__(self):
        """Unicode representation of class object."""
        name = self.staff.get_name()  # pylint: disable=no-member
        return _(f"{name}: {self.date} from {self.start} to {self.end}")

    def get_duration(self):
        """Get duration."""
        start = datetime.combine(self.date, self.start)
        end = datetime.combine(self.date, self.end)
        return end - start

    @cached_property
    def duration(self):
        """Get duration as a property."""
        return self.get_duration()


class AnnualLeave(TimeStampedModel, models.Model):
    """
    Model to keep track of staff employee annual leave.

    This model is meant to be populated once a year
    Each staff member can only have one record per leave_type per year
    """

    YEAR_CHOICES = [(r, r) for r in range(2017, datetime.today().year + 10)]

    year = models.PositiveIntegerField(
        _("Year"), choices=YEAR_CHOICES, default=2017, db_index=True
    )
    staff = models.ForeignKey(
        StaffProfile, verbose_name=_("Staff Member"), on_delete=models.CASCADE
    )
    leave_type = models.CharField(
        _("Type"), max_length=1, choices=Leave.TYPE_CHOICES, db_index=True
    )
    allowed_days = models.DecimalField(
        _("Allowed Leave days"),
        default=21,
        blank=True,
        decimal_places=1,
        max_digits=12,
        validators=[MinValueValidator(Decimal('0.1'))],
        help_text=_("Number of leave days allowed in a year."),
    )
    carried_over_days = models.DecimalField(
        _("Carried Over Leave days"),
        default=0,
        blank=True,
        decimal_places=1,
        max_digits=12,
        validators=[MinValueValidator(Decimal('0'))],
        help_text=_("Number of leave days carried over into this year."),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """Meta options for AnnualLeave."""

        verbose_name = _("Annual Leave")
        verbose_name_plural = _("Annual Leave")
        ordering = ["-year", "leave_type", "staff"]
        unique_together = (("year", "staff", "leave_type"),)

    def __str__(self):
        """Unicode representation of class object."""
        # pylint: disable=no-member
        return _(
            f"{self.year}: {self.staff.get_name()} " f"{self.get_leave_type_display()}"
        )

    def get_cumulative_leave_taken(self):
        """
        Get the cumulative leave taken.

        Returns a timedelta
        """
        return get_taken_leave_days(
            staffprofile=self.staff,
            status=Leave.APPROVED,
            leave_type=self.leave_type,
            start_year=self.year,
            end_year=self.year,
        )

    def get_available_leave_days(self, month: int = 12):
        """Get the remaining leave days."""
        if month <= 0:
            month = 1
        elif month > 12:
            month = 12

        # the max allowed days per year
        allowed = self.allowed_days

        # the days `earned` per month
        per_month = Decimal(allowed / 12)

        # the days earned so far, given the month
        earned = Decimal(month) * per_month

        # the days taken
        taken = self.get_cumulative_leave_taken()

        # the starting balance
        starting_balance = self.carried_over_days

        return Decimal(earned + starting_balance - taken)


class FreeDay(models.Model):
    """Model definition for FreeDay."""

    name = models.CharField(_("Name"), max_length=255)
    date = models.DateField(_("Date"), unique=True)

    class Meta:
        """Meta definition for FreeDay."""

        ordering = ["-date"]
        verbose_name = _("Free Day")
        verbose_name_plural = _("Free Days")

    def __str__(self):
        """Unicode representation of class object."""
        return f"{self.date.year} - {self.name}"


def get_days(start: object, end: object):
    """Yield the days between two datetime objects."""
    current_tz = timezone.get_current_timezone()
    local_start = current_tz.normalize(start)
    local_end = current_tz.normalize(end)
    span = local_end.date() - local_start.date()
    for i in range(span.days + 1):
        yield local_start.date() + timedelta(days=i)


def get_real_leave_duration(leave_obj: Leave) -> Decimal:
    """
    Get the real leave duration.

    Takes into account public holidays, weekends and weekend policy
    """
    count = Decimal(0)
    free_days = FreeDay.objects.filter(
        date__gte=leave_obj.start.date(), date__lte=leave_obj.end.date()
    ).values_list("date", flat=True)
    days = get_days(start=leave_obj.start, end=leave_obj.end)
    for day in days:
        if day not in free_days:
            day_value = settings.SSHR_DAY_LEAVE_VALUES[day.isoweekday()]
            count = count + Decimal(day_value)
    return count


def get_taken_leave_days(
    staffprofile: object, status: str, leave_type: str, start_year: int, end_year: int
):
    """
    Calculate the number of leave days actually taken.

    Takes into account weekends and weekend policy
    """
    count = Decimal(0)
    free_days = FreeDay.objects.filter(
        date__year__gte=start_year, date__year__lte=end_year
    ).values_list("date", flat=True)
    queryset = Leave.objects.filter(
        staff=staffprofile, review_status=status, leave_type=leave_type
    ).filter(Q(start__year__gte=start_year) | Q(end__year__lte=end_year))
    for leave_obj in queryset:
        days = get_days(start=leave_obj.start, end=leave_obj.end)
        for day in days:
            if day.year >= start_year and day.year <= end_year and day not in free_days:
                day_value = settings.SSHR_DAY_LEAVE_VALUES[day.isoweekday()]
                count = count + Decimal(day_value)
    return count
